from .logger import logger, debug, info, error, exception, level
from .net import request, get_ip_adr
from .shell import shell, bash, pipe
from .ssh import SSH, ParamSSH
from .dico import Dico
from .config import Config, get_app_config
from .tracer import Tracer, trace, set_tracer
from .fuzzy import fuzzy_select
from .func import singleton
from .rex import Rex, Patterns, Parser
from .utils import Flags
from .files import find, grep



from pathlib import Path

from pybrary.command import Param, ValidationError


class ParamInt(Param):
    '''Int
    '''
    def verify(self, value):
        try:
            return int(value)
        except Exception as x:
            raise ValidationError(f'{value}') from x


class ParamPath(Param):
    '''Path
    '''
    def __init__(self, *a, **k):
        self.create = k.pop('create', False)
        super().__init__(*a, **k)


class ParamFile(ParamPath):
    '''File
    '''
    def verify(self, value):
        try:
            path = Path(value)
        except Exception as x:
            raise ValidationError(f'{value}') from x
        if path.is_file():
            return path
        elif self.create:
            try:
                path.open('x')
            except Exception as x:
                raise ValidationError(f'{value}') from x
        else:
            raise ValidationError(f'"{self.name}" File not found : {value}')


class ParamDir(ParamPath):
    '''Dir
    '''
    def verify(self, value):
        try:
            path = Path(value)
        except Exception as x:
            raise ValidationError(f'\n{value}') from x
        if path.is_dir():
            return path
        elif self.create:
            try:
                path.mkdir(parents=True)
            except Exception as x:
                raise ValidationError(f'{value}') from x
        elif path.exists():
            raise ValidationError(f'"{self.name}" : {value} is not a dir')
        else:
            raise ValidationError(f'"{self.name}" dir not found : {value}')



from logging import getLogger
from os import remove, getcwd
from os.path import isfile
from shutil import copy
from functools import partial
from collections import deque

from pybrary import Rex, Patterns
from pybrary.files import  walk, split


logger = getLogger('pybrary.refiles')
debug = logger.debug


class Matcher:
    def __init__(self, root=None, full=None, path=None, name=None, ext=None, cont=None, **opts):
        self.root = root
        self.full = Patterns(full, **opts)
        self.path = Patterns(path, **opts)
        self.name = Patterns(name, **opts)
        self.ext = Patterns(ext, **opts)
        self.cont = Patterns(cont, **opts)

    def check_byline(self, full):
        excls = [p.isin for p in self.cont.excl]
        incls = [p.isin for p in self.cont.incl]
        for line in open(full):
            if any(excl(line) for excl in excls): return False
            incls = [incl for incl in incls if not incl(line)]
        return not incls

    def match(self, full, byline=True):
        full = str(full)
        path, name, ext = split(full)
        if not self.ext.find(ext): return False
        if not self.name.find(name): return False
        if not self.path.find(path): return False
        if not self.full.find(full): return False
        if self.cont:
            try:
                if byline:
                    return self.check_byline(full)
                else:
                    return self.cont.find(open(full).read())
            except Exception as x:
                debug('match : %s ! %s', full, x)
                return False
        return True

    def find(self, root=None, recur=True, rel=True):
        return find(root, matcher=self, byline=True, recur=recur, rel=rel)

    def count(self, root=None):
        return count(root, matcher=self)

    def grep(self, fil, before=0, after=0):
        return grep(fil, before=before, after=after, matcher=self)

    def greper(self, fil, before=0, after=0, color=7):
        return greper(fil, before, after, matcher=self, color=color)

    def replace(self, fil, src, dst, **opt):
        replace(fil, src, dst, matcher=self, **opt)

    def __str__(self):
        return '\n'.join(p for p in [
            '\tfull : %s' % self.full if self.full else '',
            '\tpath : %s' % self.path if self.path else '',
            '\tname : %s' % self.name if self.name else '',
            '\text : %s' % self.ext if self.ext else '',
            '\tcont : %s' % self.cont if self.cont else '',
        ] if p)


def find(root=None, ext=None, name=None, path=None, full=None, cont=None, matcher=None, byline=True, recur=True, rel=True, **opts):
    matcher = matcher or Matcher(root, ext, name, path, full, cont, **opts)
    match = partial(matcher.match, byline=byline)
    walker = walk if recur else partial(walk, recur=0)
    root = root or matcher.root or getcwd()
    len_root = len(str(root))+1
    for fil in walker(root):
        if match(fil):
            fil = str(fil)
            yield fil[len_root:] if rel else fil


def count(**opts):
    return len(list(find(**opts)))


def _grep1(lines, matcher):
    find = matcher.cont.find
    for i, line in enumerate(lines, start=1):
        if find(line):
            yield i, line


def _grep2a(lines, before, after, matcher):
    size = before*2 + 1 + after*2
    current = before*2
    current_range = list(range(current-before, current+after+1))
    find = matcher.cont.find
    dq = deque(maxlen=size)
    done = set()

    def check(line, plage=current_range):
        if find(line):
            for k in plage:
                idx, line = dq[k]
                if idx in done: continue
                yield idx, line
                done.add(idx)

    for i, src in enumerate(lines, start=1):
        dq.append((i, src))
        if len(dq) < size: continue

        # first lines
        if dq[0][0] < current:
            for n, (idx, line) in enumerate(dq):
                if idx > current: break
                plage = range(max(n-before, 0), min(n+after+1, len(dq)))
                yield from check(line, plage)

        # most lines
        idx, line = dq[current]
        yield from check(line)

    # last lines
    for n, (idx, line) in enumerate(dq):
        plage = range(max(n-before, 0), min(n+after+1, len(dq)))
        yield from check(line, plage)


def _grep2b(lines, before, after, matcher):
    check = matcher.cont.find
    found = [
        nb
        for nb, line in enumerate(lines)
        if check(line)
    ]
    next_one, last_one = 0, len(lines)-1
    for nb in found:
        first = max(nb-before, next_one)
        last = min(nb+after, last_one)
        for i in range(first, last+1):
            yield i+1, lines[i]
            next_one = i+1


def grep(fil, cont=None, before=0, after=0, matcher=None, **opts):
    matcher = matcher or Matcher(cont=cont, **opts)
    lines = open(fil).read().split('\n')
    if before or after:
        return _grep2b(lines, before, after+1, matcher)
    else:
        return _grep1(lines, matcher)


def replace(fil, src, dst, cont=None, matcher=None, count=0, dry=True, bak=True, **opts):
    if not matcher:
        cont = cont or src
        matcher = Matcher(cont=cont, **opts)
    findall = matcher.cont.findall
    src = Rex(src)
    if dry:
        for i, lig in enumerate(open(fil), start=1):
            if findall(lig):
                lig=lig.strip()
                print('\t%i\t'%i, lig)
                print('\t   ->\t', src.replace(dst, lig, count))
    else:
        bakfil = fil+'_BAK'
        copy(fil, bakfil)
        w = open(fil, 'w')
        for lig in open(bakfil):
            if findall(lig):
                w.write(src.replace(dst, lig, count))
            else:
                w.write(lig)
        w.close()
        if not bak: remove(bakfil)


def color(txt, col=7):
    if col>7:
        return txt
    col+=30
    esc = chr(27)
    return f'{esc}[1;{col}m{txt}{esc}[0m'


def greper(fil, before=0, after=0, matcher=None, **opts):
    assert isfile(fil), '%s is not a file'%fil
    col = opts.pop('color', 7)
    if not matcher:
        matcher = Matcher(**opts)
    rex = matcher.cont.incl[0]
    for nb, line in matcher.grep(fil, before, after):
        for grp in rex.findall(line):
            line = line.replace(grp, color(grp, col))
        yield f'\t{nb:3d}\t{line}'




from os import environ
from pathlib import Path
from subprocess import call
from warnings import deprecated

from pybrary import Dico


def escape(value):
    if isinstance(value, str):
        return value.replace('\\', '\\\\')
    else:
        return value


def create_config_py(path, defaults):
    kv = [(k, f"'{v}'" if isinstance(v, str) else v) for k, v in defaults.items()]
    kv = ',\n    '.join(f'{k} = {escape(v)}' for k, v in kv)
    with open(path, 'w') as out:
        out.write(f'config = dict(\n    {kv}\n)\n')

def create_config_toml(path, defaults):
    from toml import dump
    content = dict(config = defaults)
    with open(path, 'w') as out:
        dump(content, out)

def create_config_yaml(path, defaults):
    from yaml import dump
    content = dict(config = defaults)
    with open(path, 'w') as out:
        dump(content, out, indent=4)

def create_config_json(path, defaults):
    from json import dump
    content = dict(config = defaults)
    with open(path, 'w') as out:
        dump(content, out, indent=4)

creators = dict(
    py = create_config_py,
    toml = create_config_toml,
    yaml = create_config_yaml,
    json = create_config_json,
)


def load_config_py(path):
    from pybrary.modules import load
    return load('config', path).config

def load_config_toml(path):
    from tomllib import loads
    return loads(path.read_text())['config']

def load_config_yaml(path):
    from yaml import load, SafeLoader
    return load(open(path), Loader=SafeLoader)['config']

def load_config_json(path):
    from json import load
    return load(open(path))['config']

loaders = dict(
    py = load_config_py,
    toml = load_config_toml,
    yaml = load_config_yaml,
    json = load_config_json,
)


class Config:
    def __init__(self, app, defaults=None, ext='py'):
        self.app = app
        defaults = defaults or dict()
        found = self.find() or self.create(defaults, ext)
        self.config = Dico(found)

    def get_env(self, name):
        # NOTE : all value in environ are str
        return environ.get(f'{self.app}_{name}')

    def __getattr__(self, name):
        return self.get_env(name)  or self.config.get(name)

    @property
    def root(self):
        path = Path(f'~/.config/{self.app}').expanduser()
        path.mkdir(mode=0o700, parents=True, exist_ok=True)
        return path

    def find(self):
        for ext, loader in loaders.items():
            full = self.root / f'config.{ext}'
            if full.is_file():
                config = loader(full)
                self.path = full
                return config

    def create(self, defaults, ext):
        full = self.root / f'config.{ext}'
        creators[ext](full, defaults)
        loader = loaders[ext]
        config = loader(full)
        self.path = full
        return config

    def edit(self):
        editor = environ.get('EDITOR', 'vim')
        call([editor, self.path])


@deprecated('Use pybrary.Config instaed')
def get_app_config(app):
    path = Path(f'~/.config/{app}').expanduser()
    try:
        from pybrary.modules import load
        full = path / 'config.py'
        config = load('config', full)
        return full, config.config
    except: pass
    try:
        from tomllib import loads
        full = path / 'config.toml'
        config = loads(full.read_text())
        return full, config
    except: pass
    try:
        from yaml import load, SafeLoader
        full = path / 'config.yaml'
        config = load(full, loader=SafeLoader)
        return full, config
    except: pass
    try:
        from json import load
        full = path / 'config.json'
        config = load(full)
        return full, config
    except: pass
    return None, None



from functools import wraps
from inspect import (
    _empty,
    currentframe,
    getargvalues,
    getouterframes,
    signature,
)
from opcode import opname
from os import getcwd
from re import compile
from sys import settrace, path
from time import perf_counter_ns


lib_paths = (
    '_bootstr',
    '<.*>',
    'site-packages/.*',
)

lib_func = (
    '<.*>',
)


class Tracer:
    def __init__(self,
        label         = 'tracing',
        full_path     = False,
        show_types    = False,
        check_types   = False,
        show_times    = False,
        show_stamp    = False,
        stdlib        = False,
        format_arg_names = None,
        format_arg_types = None,
        dest          = None,
        buffered      = True,
        echo          = True,
        quiet         = True,
        ignored_funcs = None,
        ignored_paths = None,
    ):
        self.label = label
        self.level = 0
        self.offbak = list()
        self.full_path = full_path
        self.traces = list()
        self.show_types = show_types
        self.check_types = check_types
        self.ignored_paths = set()
        self.ignored_funcs = set()
        self.truncated = list()
        self.show_times = show_times
        self.show_stamp = show_stamp
        self.dest = dest or getcwd()
        self.buffered = buffered
        self.echo = echo
        self.quiet = quiet
        self.stdlib = stdlib
        self.setup(ignored_paths, ignored_funcs)
        self.format_arg_names = dict(
        )
        if format_arg_names:
            self.format_arg_names.update(format_arg_names)
        self.format_arg_types = {
            'list': self.format_arg_list,
            'dict': self.format_arg_dict,
            'function': self.format_arg_function,
            'generator': self.format_arg_generator,
        }
        if format_arg_types:
            self.format_arg_types.update(format_arg_types)


    @property
    def output(self):
        return f'{self.dest}/{self.label}.trace'

    def create(self):
        try:
            with open(self.output, 'x'):
                self.log(f'>>> {self.label}', ts=False)
        except FileExistsError: pass

    def write(self, line):
        with open(self.output, 'a') as out:
            out.write(f'{line}\n')

    def log(self, msg, ts=True):
        ts = f'{self.timestamp()}@{self.label[:12]:<12}' if ts and self.show_stamp else ''
        tab = '    ' * self.level
        msg = f'{ts}{tab}{msg}'
        if self.buffered:
            self.traces.append(msg)
        else:
            self.write(msg)
        if self.echo:
            print(msg)

    def log_exc(self, exc, info=None):
        if not self.quiet:
            msg = f'! {type(exc).__name__} ! {exc}'
            if info:
                msg = f'! {info} {msg}'
            self.log(msg)

    def dump(self):
        self.create()
        with open(self.output, 'a') as out:
            out.write('\n'.join(self.traces))

    def truncate(self, path):
        if not self.full_path:
            for pref in self.truncated:
                if path.startswith(pref):
                    path = path[len(pref)+1:]
        if path.endswith('.py'):
            path = path[:-3]
        return path

    def ignore_path(self, path):
        return self.ignored_paths.add(compile(path).search)

    def ignored_path(self, path):
        return any(ignore(path) for ignore in self.ignored_paths)

    def ignore_func(self, func):
        return self.ignored_funcs.add(compile(func).search)

    def ignored_func(self, func):
        return any(ignore(func) for ignore in self.ignored_funcs)

    def truncate_path(self, prefix):
        self.truncated.append(prefix)

    def timestamp(self):
        return perf_counter_ns()

    def start_timer(self):
        if self.show_times:
            self.timers[self.level] = self.timestamp()

    def stop_timer(self):
        if self.show_times:
            self.timers[self.level] = self.timestamp() - self.timers[self.level]

    @property
    def elapsed(self):
        if self.show_times:
            elapsed = self.timers[self.level] // 1_000_000
            if elapsed < 10000:
                return f' [{elapsed} ms]' if elapsed > 0 else ''
            else:
                return f' [{elapsed // 1000} s]'
        else:
            return ''

    def setup(self, ignored_paths, ignored_funcs):
        std_paths = [p for p in path if 'python' in p]
        std_paths.extend(lib_paths)
        if not self.stdlib:
            for p in std_paths:
                self.ignore_path(p)
        if ignored_paths:
            for p in ignored_paths:
                self.ignore_path(p)
        if not self.full_path:
            self.truncated.extend(std_paths)
        for func in lib_func:
            self.ignore_func(func)
        if ignored_funcs:
            for f in ignored_funcs:
                self.ignore_func(f)
        self.ignore_func(r'Tracer\..*')
        self.ignore_path('tracer')
        self.truncate_path(getcwd())
        self._exc_ = False
        self.timers = dict()

    def format_arg_type(self, val):
        return self.format_arg_types.get(type(val).__name__, str)(val)

    def format_arg_return(self, val):
        s = '' if val is None else self.format_arg_type(val)
        if self.show_types:
            s = f':{type(val).__name__} = {s}'
        return s

    def format_arg_name(self, name, val):
        formatter = self.format_arg_names.get(name)
        if formatter: return formatter(val)

    def format_arg(self, k, v):
        try:
            f = self.format_arg_name(k, v)
            if f is None:
                f = self.format_arg_type(v)
        except Exception as x:
            self.log_exc(x, f'{k}: {type(v).__name__}')
            f = ' ! '
        return f

    def format_arg_dict(self, val):
        return ', '.join(f'{k}:{self.format_arg(k, v)}' for k, v in val.items())

    def format_arg_list(self, val):
        return ', '.join(self.format_arg(i, v) for i, v in enumerate(val))

    def format_arg_function(self, func):
        mname = func.__module__
        mname = f'{mname}.' if mname!= '__main__' else ''
        fname = func.__qualname__.replace('.<locals>', '')
        return mname + fname

    def format_arg_generator(self, gen):
        gname = gen.__qualname__.replace('.<locals>', '')
        return gname

    def format_args(self, args):
        if self.show_types:
            return [f'{a if isinstance(a, str) else str(a)}:{type(a).__name__}' for a in args]
        else:
            return [a if isinstance(a, str) else str(a) for a in args]

    def format_kws(self, args):
        items = list()
        for k, v in args.items():
            f = self.format_arg(k, v)
            if self.show_types:
                t = type(args[k]).__name__
                s = f'{k}:{t} = {f}'
            else:
                s = f'{k} = {f}'
            items.append(s)
        return items

    def trace(self, file, line, func, args):
        if isinstance(args, tuple):
            if isinstance(args[0], str):
                pfx = args[0]
                if pfx == '>>>':
                    val = self.format_arg_return(args[1])
                    args = f' -> {val}{self.elapsed}'
                elif pfx == '-->':
                    val = self.format_arg_return(args[1])
                    args = f' >> {val}{self.elapsed}'
                elif pfx == '!!!':
                    args = f'{file}.{func}:{line}\n! {args[1].__name__} ! {args[2]}'
                else:
                    self.log(f'???2 {args}')
            elif len(args)==2 and isinstance(args[0], tuple) and isinstance(args[1], dict):
                args, kws = args
                args = ', '.join(a for a in self.format_args(args)) if args else ''
                kws = ', '.join(a for a in self.format_kws(kws)) if kws else ''
                sep = ', ' if args and kws else ''
                args = '('+ args + sep + kws +')'
            else:
                self.log(f'???3 {args}')
        else:
            self.log(f'???4 {args}')
        self.log(f'{file}:{line} {func}{args}')

    def ann_err(self, name, expected, actual):
        if name == '_return_':
            msg = f'! "should return "{expected.__name__}" not "{actual.__name__}" !'
        else:
            msg = f'! arg "{name}" should be "{expected.__name__}" not "{actual.__name__}" !'
        self.log(msg)

    def get_func_args(self, frame):
        arginfo = getargvalues(frame)
        args = arginfo.locals[arginfo.varargs] if arginfo.varargs else tuple()
        kws = arginfo.locals[arginfo.keywords] if arginfo.keywords else dict()
        return args, kws

    def get_sig(self, frame):
        fct = frame.f_code.co_name
        if fct.endswith('comp>'): return
        slf = frame.f_locals.get('self')
        if slf:
            # method
            func = getattr(slf, fct)
        else:
            try:
                # global function
                func = frame.f_globals[fct]
            except KeyError:
                # inner function
                outer = getouterframes(frame)[1]
                func = outer.frame.f_locals[fct]
        try:
            sig = signature(func)
        except Exception as x:
            sig = None
        return sig

    def chk_arg_ann(self, name, arg, ann):
        if ann is _empty: return
        if ann is None and arg is None: return
        try:
            if not isinstance(arg, ann):
                self.ann_err(name, ann, type(arg))
        except TypeError:
            self.log(f'! unable to  check arg ! {arg}:{ann}')

    def check_sig(self, frame, args):
        if self.check_types:
            sig = self.get_sig(frame)
            if not sig or not sig.parameters: return
            if isinstance(args, tuple) and args[0] == '>>>':
                arg = args[1]
                ann = sig.return_annotation
                self.chk_arg_ann('_return_', arg, ann)
            elif isinstance(args, dict):
                for name, p in sig.parameters.items():
                    arg = args[name]
                    ann = p.annotation
                    if ann is _empty: continue
                    self.chk_arg_ann(name, arg, ann)

    def get_func_name(self, frame):
        fct = frame.f_code.co_name
        slf = frame.f_locals.get('self')
        try:
            repr(slf) # check handlability
        except Exception:
            slf = None
        if slf:
            try:
                func = getattr(slf, fct)
            except Exception:
                func = None
        else:
            func = frame.f_globals.get(fct)
        try:
            return func.__qualname__
        except Exception:
            return fct

    def get_codename(self, frame):
        codeval = frame.f_code.co_code[frame.f_lasti]
        if not isinstance(codeval, int):
            codeval = ord(codeval)
        codename = opname[codeval]
        return codename

    def get_trace(self, frame, arg, lvl=0):
        path = frame.f_code.co_filename
        if self.ignored_path(path): return False
        func = self.get_func_name(frame)
        if self.ignored_func(func): return False
        self.level += lvl
        self.check_sig(frame, arg)
        path = self.truncate(path)
        line = frame.f_lineno
        self.trace(path, line, func, arg)
        return path, line, func, arg

    def trace_exception(self, frame, typ, val, tbk):
        ret = self.get_trace(frame, ('!!!', typ, val))
        if ret:
            self._exc_ = True

    def trace_return(self, frame, val):
        self.stop_timer()
        if self._exc_:
            self.level-=1
            self._exc_ = False
            return

        retname = self.get_codename(frame)
        if retname in ('RETURN_VALUE', 'RETURN_CONST'):
            pfx = '>>>'
        elif retname in ('YIELD_VALUE', 'YIELD_CONST'):
            pfx = '-->'
        elif retname in ('RAISE_VARARGS', 'RERAISE') and val is None:
            # return from __init__
            return
        else:
            # self.log(f'???1 {retname} {val}')
            pfx = retname

        if self.get_trace(frame, (pfx, val)):
            self.level-=1

    def trace_call(self, frame, event, arg):
        codename = self.get_codename(frame)
        if codename == 'YIELD_VALUE':
            # don't trace "yield from" calls, trace inner generator only
            return
        args = self.get_func_args(frame)
        self.get_trace(frame, args, 1)
        self.start_timer()
        return self.trace_event

    def trace_event(self, frame, event, arg):
        if event == 'call':
            return self.trace_call

        if event == 'return':
            self.trace_return(frame, arg)

        if event == 'exception':
            self.trace_exception(frame, *arg)

    def __enter__(self):
        if not self.buffered: self.create()
        self.log(f'{self.label} >>>', ts=False)
        self.offbak.append(self.level)
        self.level += 1
        if self.level == 1:
            settrace(self.trace_event)

    def __exit__(self, typ, val, tbk):
        self.level = self.offbak.pop()
        if self.level == 0:
            settrace(None)
        self.log(f'{self.label} <<<', ts=False)
        if self.buffered: self.dump()
        return self.quiet


_tracer_ = Tracer(
    label = 'trace',
    quiet = False,
)

def set_tracer(tracer=None, **k):
    global _tracer_
    if isinstance(tracer, Tracer):
        _tracer_ = tracer
    else:
        _tracer_ = Tracer(**k)

def trace(fct):
    @wraps(fct)
    def wrapper(*a, **k):
        caller = getouterframes(currentframe())[1]
        with _tracer_:
            _tracer_.log(f'< {_tracer_.truncate(caller.filename)}:{caller.lineno} {caller.function} > {caller.code_context[0].strip()}')
            _tracer_.level -= 1
            try:
                result = fct(*a, **k)
            except Exception as x:
                _tracer_.level += 1
                _tracer_.log_exc(x)
                result = None
        return result
    return wrapper



from http import HTTPStatus
from http.server import BaseHTTPRequestHandler, HTTPServer

from contextlib import contextmanager
from functools import partial
from threading import Thread
from time import sleep
from urllib.request import urlopen, Request


class RequestHandler(BaseHTTPRequestHandler):
    routes = (
        ('get', '/hello', 'hello'),
    )

    def __init__(self, *a, **k):
        self.map = dict(
            get  = dict(),
            post = dict(),
        )
        for meth, path, func in self.routes:
            self.map[meth][path] = getattr(self, func)
        self.map['post']['/shutdown'] = self.shutdown
        super().__init__(*a, **k)

    def hello(self):
        return HTTPStatus.OK, 'Hi there'

    def shutdown(self):
        setattr(self.server, '_BaseServer__shutdown_request', True)
        return 200, 'shuting down'

    def not_found(self):
        return HTTPStatus.NOT_FOUND, f'{self.command} {self.path} Not Found'

    def response(self, code, body):
        self.send_response(code)
        self.send_header('Content-type', 'text/html')
        self.end_headers()
        self.wfile.write(body.encode())

    @property
    def body(self):
        size = int(self.headers['Content-Length'])
        return self.rfile.read(size).decode()

    def do_GET(self):
        print(f'GET {self.path}\n{self.headers}')
        route = self.map['get'].get(self.path, self.not_found)
        code, body = route()
        self.response(code, body)

    def do_POST(self):
        print(f'POST {self.path}\n{self.headers}\n{self.body}')
        route = self.map['post'].get(self.path, self.not_found)
        code, body = route()
        self.response(code, body)


class Server:
    '''Generic HTTP server.

    Intended to be used as a mock for testing puposes.
    '''
    def __init__(self, Handler=RequestHandler):
        self.Handler = Handler

    def serve(self, adr='localhost', port=7777):
        server = HTTPServer((adr, port), self.Handler)
        print(f'Listening on {adr}:{port} ...\n')
        try:
            server.serve_forever()
        except KeyboardInterrupt:
            pass
        server.server_close()


def request(url, method='GET'):
    resp = urlopen(Request(url, method=method))
    code = resp.code
    text = resp.read().decode('utf-8')
    return code, text


@contextmanager
def serve(Handler, adr='localhost', port=7777):
    server = Server(Handler)
    thread = Thread(
        target = partial(server.serve, adr=adr, port=port),
        daemon = True,
    )
    thread.start()
    sleep(1)

    yield

    code, text = request(f'http://{adr}:{port}/shutdown', 'POST')
    assert code == 200
    thread.join()



from functools import partial
from datetime import datetime
from io import SEEK_END
from operator import attrgetter
import os, os.path, stat
from os import scandir, getcwd
from os.path import splitext, splitdrive, isfile, abspath
from pathlib import Path
from re import compile, match


def walk(path=None, recur=99):
    path = Path(path).expanduser()
    with scandir(path or getcwd()) as ls:
        for entry in sorted(ls, key=attrgetter('path')):
            if entry.is_dir(follow_symlinks=False):
                if recur:
                    yield from walk(entry.path, recur-1)
            elif entry.is_file():
                yield Path(entry)


def find(path, name):
    search = compile(name).search
    for path in walk(path):
        if search(str(path)):
            yield path


def grep(path, rex):
    search = compile(rex).search
    for line in open(path):
        if found := search(line):
            if groups := found.groupdict():
                yield groups
            elif groups := found.groups():
                yield groups
            else:
                yield line.rstrip()


class FileStat:
    def __init__(self, path):
        self.path = path if isinstance(path, Path) else Path(path)
        self.stat = os.stat(str(self.path))

    @property
    def mtime(self):
        mtime = self.stat[stat.ST_MTIME]
        return datetime.fromtimestamp(mtime)

    @property
    def age(self):
        delta = datetime.now() - self.mtime
        return delta.days


def last_lines(path, nb_lines, max_line_len=99):
    offset = nb_lines * max_line_len
    with open(path, 'rb') as inp:
        try:
            inp.seek(-offset, SEEK_END)
        except OSError:
            inp.seek(0)
        raw = inp.read()
    lines = raw.decode().split('\n')
    last = lines[-nb_lines:]
    return last


def split(full):
    path, name = os.path.split(full)
    drive, path = splitdrive(path)
    name, ext = splitext(name)
    ext = ext[1:]
    return path, name, ext



from re import escape, compile


def fuzzy_select(pattern, inputs, protector=None, decorate=None):
    '''Fuzzy selector.

    Returns a weigthed and sorted selection of items
    from (decorated) inputs based on "fuzzy" pattern.

    pattern   :
        A string representing the pattern to be found "fuzzily".

    inputs    :
        An iterable of strings to be filtered.

    protector :
        The string used as a delimitor of "protected" sections in pattern.
        defaults to ".

    decorate  :
        An optional callable to apply on each input before filtering.

    returns   :
        A generator of selectioned inputs.
    '''

    # define local variables.
    parts, pro, deco = [], protector or '"', decorate

    # make pattern spliter.
    split = compile('({pro}.*?{pro})'.format(pro=pro)).split

    # define local functions
    app, ext, pat = parts.append, parts.extend, split(pattern)

    # extract sub-patterns from pattern.
    [app(p.strip(pro)) if pro in p else ext(tuple(p.strip())) for p in pat]

    # build pattern finder.
    search = compile('.*?'.join(map(escape, parts))).search

    # decorate inputs if needed.
    decorated = map(deco, inputs) if callable(deco) else inputs

    # select matching inputs.
    matching = ((i, m) for i, m in enumerate(map(search, decorated)) if m)

    # apply weigths to matchings
    weighted = ((len(m.group()), m.start(), i) for i, m in matching)

    # sort selection.
    suggested = [inputs[w[-1]] for w in sorted(weighted)]

    return suggested



class Primes:
    def __init__(self):
        self.primes = [3]
        '''
        self.primes = array('L', [3]) # ! bigger and slower !
        '''

    def next(self):
        primes = self.primes
        n = primes[-1]
        while True:
            n+=2
            for i in primes:
                if n%i==0: break
            else:
                primes.append(n)
                return n

    def __getitem__(self, i):
        if i<0: raise OverflowError
        if i==0: return 2
        primes = self.primes
        nxt = self.next
        while len(primes)<i: nxt()
        return primes[i-1]

    def __iter__(self):
        yield 2
        for p in self.primes: yield p
        while True: yield self.next()

    def __contains__(self, num):
        if num>2:
            if num%2==0: return False
        else:
            return num==2
        last = int(num**.5)
        for p in self:
            if num==p: return True
            if num%p==0: return False
            if p>last: return True

    def factors(self, num):
        if num<2: return []
        last = int(num**.5)
        for p in self:
            if num==p: return [num]
            if num % p == 0: return [p] + self.factors(num//p)
            if p>last: return [num]


primes = Primes()




from collections import defaultdict
from pathlib import Path


def get_makefile():
    for name in (
        'Makefile',
        'makefile',
    ):
        makefile = Path(name)
        if makefile.is_file():
            return makefile
    raise RuntimeError(' ! Makefile not found')


def index(makefile=None):
    makefile = makefile or get_makefile()
    targets = dict()
    for n, line in enumerate(open(makefile)):
        first, line = ord(line[0]), line.rstrip()
        if first == 9:
            pass
        elif first == 10:
            end = n-1
            targets[target] = start, end
        else:
            target, _, prereq = line.partition(':')
            start = n
    return targets


def parse(makefile=None):
    makefile = makefile or get_makefile()
    targets = defaultdict(list)
    for line in open(makefile):
        first, line = ord(line[0]), line.rstrip()
        if first == 9:
            targets[target].append(line)
        elif first == 10:
            pass
        else:
            target, _, prereq = line.partition(':')
            if pre := prereq.strip():
                targets[target].append(pre)
    return targets


def show(makefile=None):
    makefile = makefile or get_makefile()
    for target, commands in parse(makefile).items():
        print(f'{target}:')
        for cmd in commands:
            print(f'    {cmd}')
        print()




from logging import getLogger, NullHandler, DEBUG, INFO, ERROR


logger = getLogger('Pybrary')
logger.addHandler(NullHandler())
debug = logger.debug
info = logger.info
error = logger.error
exception = logger.exception


def level(level):
    if level.lower()=='debug':
        level = DEBUG
    elif level.lower()=='info':
        level = INFO
    elif level.lower()=='error':
        level = ERROR
    else:
        raise ValueError('level must be in debug, info, error')
    logger.level = level



import pybrary


def get_ip_adr():
    try:
        ok, adr = pybrary.get_ip_adr()
        if ok:
            return True, adr
        else:
            return False, f'Error ! {adr}'
    except Exception as x:
        return False, f'Exception ! {x}'


def bash(script):
    ret, out, err = pybrary.bash(script)
    ok = ret==0
    return ok, out if ok else f'\n > {out}\n ! {err}\n'


def ssh(host, cmd):
    ssh = pybrary.SSH()
    rem = ssh.hosts[host]
    ret, out, err = rem.run(cmd)
    return ret==0, out.strip()


def known_hosts():
    ssh = pybrary.SSH()
    ssh.hosts.known()
    return True, ''


def find(root='.', pattern=r'.+\.py$'):
    for path in pybrary.find(root, pattern):
        print(path)
    return True, ''


def grep(pattern, root='.', files=r'.+\.py$'):
    for path in pybrary.find(root, files):
        header, footer = True, False
        try:
            for line in pybrary.grep(path, pattern):
                if header:
                    print(path)
                    header = False
                    footer = True
                print(f'\t{line}')
        except Exception as x:
            if header:
                print(path)
                header = False
            print(f' ! {x}')
            footer = True
        if footer:
            print()
    return True, ''




import re


def negates(fct):
    def deco(*a, **k): return not fct(*a, **k)
    return deco


class Rex:
    def __init__(self, pat, match=True, switch=None, **opts):
        flags = re.IGNORECASE
        if opts.get('case', False): flags -= re.IGNORECASE
        if '\n' in pat: flags += re.VERBOSE
        flags += re.UNICODE
        self.pat = re.compile(pat, flags)
        self.match = self._match if match else negates(self._match)
        self.isin = self._isin if match else negates(self._isin)
        if switch: self.switch = switch

    def _match(self, txt):
        return self.pat.match(txt) and True or False

    def _isin(self, txt):
        return self.pat.search(txt) and True or False

    def find(self, txt):
        m = self.pat.search(txt)
        if m:
            return m.group()

    def findall(self, txt):
        found = list()
        for groups in self.pat.findall(txt):
            if isinstance(groups, tuple):
                for i in groups:
                    found.append(i)
            else:
                found.append(groups)
        return found

    def split(self, txt):
        return [i for i in self.pat.split(txt) if i]

    def replace(self, repl, txt, count=0):
        return self.pat.sub(repl, txt, count)

    def switch(self, txt):
        # to be overwritten. called by subfct.
        return txt

    def _repl(self, match):
        return self.switch(match.group())

    def subfct(self, txt, count=0):
        return self.pat.sub(self._repl, txt, count)

    def __str__(self):
        return self.pat.pattern


class Patterns:
    def __init__(self, pattern=None, match=True, **opts):
        self.incl = []
        self.excl = []
        self.valid = False
        if pattern:
            self(pattern, match, **opts)

    def __bool__(self):
        return self.valid

    def __call__(self, pat, match=True, **opts):
        if match:
            self.incl.append(Rex(pat, match=True, **opts))
        else:
            self.excl.append(Rex(pat, match=True, **opts))
        self.valid = True

    def matchany(self, txt):
        if not self.valid: return True
        excl = any(p.match(txt) for p in self.excl) if self.excl else False
        if excl: return False
        return any(p.match(txt) for p in self.incl) if self.incl else True

    def matchall(self, txt):
        if not self.valid: return True
        excl = all(p.match(txt) for p in self.excl) if self.excl else False
        if excl: return False
        return all(p.match(txt) for p in self.incl) if self.incl else True

    def findany(self, txt):
        if not self.valid: return True
        excl = any(p.isin(txt) for p in self.excl) if self.excl else False
        if excl: return False
        return any(p.isin(txt) for p in self.incl) if self.incl else True

    def findall(self, txt):
        if not self.valid: return True
        excl = all(p.isin(txt) for p in self.excl) if self.excl else False
        if excl: return False
        return all(p.isin(txt) for p in self.incl) if self.incl else True

    def find(self, txt):
        if any(p.isin(txt) for p in self.excl): return False
        return all(p.isin(txt) for p in self.incl)

    def __str__(self):
        return '{} - {}'.format(', '.join(map(str, self.incl)), ', '.join(map(str, self.excl)))


class Parser:
    def __init__(self, spec):
        tokens = self.make(spec)
        self.scan = re.Scanner(tokens).scan

    def make(self, spec):
        def handle(name):
            if name:
                return lambda scanner, token: (name, token)
        tokens = [
            (pattern, handle(name))
            for name, pattern in spec
        ]
        return tokens

    def parse(self, text):
        tokens, _ = self.scan(text)
        return tokens



from inspect import stack


class Flags:
    def __init__(self, *flags):
        self.flags = flags
        for i, f in enumerate(self.flags):
            setattr(self, f, 2**i)

    def expose(self):
        stack()[1].frame.f_globals.update(self)

    def __iter__(self):
        return ((k, getattr(self, k)) for k in self.flags)

    def __str__(self):
        return 'Flags :\n\t'+'\n\t'.join(f'{k} = {v}' for k, v in self)




from pickle import dump, load

import nicely


class Dico(dict):
    def _cast(self, value):
        if isinstance(value, dict) and not isinstance(value, Dico):
            value = Dico(value)
        return value

    def __getitem__(self, item):
        try:
            value = super().__getitem__(item)
            return self._cast(value)
        except KeyError:
            value = self[item] = Dico()
            return value

    def __getattr__(self, attr):
        if attr.startswith('_'):
            value = super().__getattr__(attr)
            return value
        else:
            value = self[attr]
            return self._cast(value)

    def __setattr__(self, attr, val):
        if attr.startswith('_'):
            super().__setattr__(attr, val)
        else:
            self[attr] = self._cast(val)

    def __getstate__(self):
        return dict(self.items())

    def dump(self, path):
        items = dict(self)
        with open(path, 'wb') as out:
            dump(items, out)

    def load(self, path):
        with open(path, 'rb') as inp:
            items = load(inp)
        self.update(items)

    def __str__(self):
        return nicely.format(self)
